/* -------------------------------------------------------------------------- */
/* Copyright 2002-2006 GridWay Team, Distributed Systems Architecture         */
/* Group, Universidad Complutense de Madrid                                   */
/*                                                                            */
/* Licensed under the Apache License, Version 2.0 (the "License"); you may    */
/* not use this file except in compliance with the License. You may obtain    */
/* a copy of the License at                                                   */
/*                                                                            */
/* http://www.apache.org/licenses/LICENSE-2.0                                 */
/*                                                                            */
/* Unless required by applicable law or agreed to in writing, software        */
/* distributed under the License is distributed on an "AS IS" BASIS,          */
/* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   */
/* See the License for the specific language governing permissions and        */
/* limitations under the License.                                             */
/* -------------------------------------------------------------------------- */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "drmaa.h"
#include <unistd.h>

void setup_job_template( drmaa_job_template_t **jt, int coorda, int coordb, float c, int n);

int main(int argc, char *argv[])
{
	char                  	error[DRMAA_ERROR_STRING_BUFFER];
	int                    	result, result2;
	drmaa_job_template_t * 	jt;
	char                   	job_id[DRMAA_JOBNAME_BUFFER];
	char                   	job_id_out[DRMAA_JOBNAME_BUFFER];    
	drmaa_attr_values_t *  	rusage;
	int                    	stat;    
	char                   	attr_value[DRMAA_ATTR_BUFFER];

	int			i,j,z,h,g; // Aux variables

	int			levels = atoi(argv[1]); // Levels of the Algorithm, to be deduced
	int			rellevel = atoi(argv[2]); // Level of Reliability, to be passed by argv
	int			agglevel = atoi(argv[4]); // Level where to start agglomeration (-1 if none)
	char*			reltype = argv[3]; // Type of Reliability			
	
	float			c = 0.9;
	int			n = 5;
	
	FILE *f1;
	
	f1 = fopen ("output.txt", "w");
	fprintf (f1,"\n");
	fclose(f1);
	
	
	// Process Type of Reliability
	// 
	//	cp = Redundancy - Only at Critical Path (BY DEFAULT) - 0
	//	al = Redundancy - All the nodes - 1
	//	ag = Agglomeration 
	
	int	reltnum = 0; 
	
	if (strcmp(reltype,"al") == 0)
	{
		reltnum = 1;
		f1 = fopen ("output.txt", "a");
		fprintf(f1,"Reliability Type: Redundancy - All Nodes\n");
		fclose(f1);
	}
	else if (strcmp(reltype,"cp") == 0)
	{
		f1 = fopen ("output.txt", "a");
		fprintf(f1,"Reliability Type: Redundancy - Critical Path\n");
		fclose(f1);
	}
	if (agglevel != -1)
	{
		f1 = fopen ("output.txt", "a");
		fprintf(f1,"Reliability Type: Agglomeration at Level %d\n",agglevel);
		fclose(f1);
	}

	// Obtain number of MetaJobs
	int mtnum = 0;
	for (i=1;i<levels;i++) mtnum = mtnum + i;
	
	f1 = fopen ("output.txt", "a");
	fprintf(f1,"Metajobs considered: %d\n", mtnum);
	fclose(f1);

	// Declare the Metajobs Matrixes
	int mtcoord[mtnum][2]; // Metajobs coordinates
	int metajob[mtnum][rellevel]; // Metajobs copies
	int mtdep[mtnum][2]; // Metajobs dependencies 
	int mtagg[2]; // Metajobs (start and end) for Agglomeration
	int jobagg[2]; // Jobs (start and end) for Agglomeration
        int aggnumjobs; // Number of jobs for agglomeration
	char aggcommand[256]; // Command executed for agglomeration
	

	// Obtain dependencies
	z = 0; // Used for level
	h = 0; // Used for that MT
	for (i=levels-1;i>0;i--)
	{
		// If Aggregation was selected
		if (z == agglevel)
		{
			mtagg[0] = h;
			aggnumjobs = i + rellevel - 1;
		}
	
		for (j=z;j<levels-1;j++)
		{	
			// Metajob coordinates
			mtcoord[h][0] = z;
			mtcoord[h][1] = j;

			// First gridified level?
			if (z==0)
			{
				// No dependencies
				mtdep[h][0] = -1;
				mtdep[h][1] = -1;
			}
			else
			{
// 				mtdep[h][0] = z-1;
// 				mtdep[h][1] = j;
				// Get MT number
				
				// 1. MT depends on CP MT [(z-1),0]
				for (g=0;g<h;g++)
				{
					if ((mtcoord[g][0] == (z - 1)) && (mtcoord[g][1] == (z - 1)))
					{
						mtdep[h][0] = g;
					}
				}
				// 2. MT depends on MT above
				for (g=0;g<h;g++)
				{
					if ((mtcoord[g][0] == (z - 1)) && (mtcoord[g][1] == j))
					{
						mtdep[h][1] = g;
					}
				}
				
			}
			
			// Initialize metajob[mtnum][rellevel]
			for (g=0;g<rellevel;g++)
			{
				metajob[h][g] = -1;
			}
			f1 = fopen ("output.txt", "a");
			fprintf (f1,"%d [%d,%d](%d,%d) ",h,mtcoord[h][0],mtcoord[h][1],mtdep[h][0],mtdep[h][1]);	
			fclose(f1);
			h++;
		}
		f1 = fopen ("output.txt", "a");
		fprintf (f1," %d\n",z);
		fclose(f1);
		z++;	
	}
	
	// Last MT for aggregation
	if (agglevel!=-1)
	{
		mtagg[1] = h - 1;
		f1 = fopen ("output.txt", "a");
		fprintf (f1,"\nAgglomeration: Metajobs %d to %d\n",mtagg[0],mtagg[1]);
		fclose(f1);
	}

						
	result = drmaa_init (NULL, error, DRMAA_ERROR_STRING_BUFFER-1);
	
	if ( result != DRMAA_ERRNO_SUCCESS)
	{
		f1 = fopen ("output.txt", "a");
		fprintf(f1,"drmaa_init() failed: %s\n", error);
		fclose(f1);
		return -1;
	}
	else
	f1 = fopen ("output.txt", "a");
	fprintf (f1,"drmaa_init() success \n");
	fclose(f1);
	
	int firstlevel = 1; // First level?
	int selfjob; // Number of Copy (not JobID)

	int mtdone[mtnum]; // If MT is finished
	int mtreleased[mtnum]; // If MT was released

	// For each MetaJob, submit copy
	for (i=0;i<mtnum;i++)
	{
		f1 = fopen ("output.txt", "a");
		fprintf(f1,"\n* MetaTask %d\n",i);
		fclose(f1);

		mtdone[i] = 0; // MT is not DONE (has yet started)

		if (mtdep[i][0] == -1) 
		{
			firstlevel = 1;
			f1 = fopen ("output.txt", "a");
			fprintf(f1,"-> FIRST LEVEL (Job won't be hold)\n");
			fclose(f1);
			mtreleased[i] = 1; // MT is released (It won't be hold)
		}
		else
		{
			firstlevel = 0;
			mtreleased[i] = 0; // MT is not released
		}

		
	
		for (selfjob=0;selfjob<rellevel;selfjob++)
		{		
			// For each job
			
			//If not from critical path, just one copy
			
			if ((reltnum == 1) || ((mtcoord[i][0] == mtcoord[i][1]) || ((mtcoord[i][0] != mtcoord[i][1]) && (selfjob==0))))
			{
			
				// CREATE JOB TEMPLATE			
				setup_job_template(&jt, mtcoord[i][1]+1, mtcoord[i][0], c, n);
				//////////////////////////////////////////////////////////////
				
				result = drmaa_run_job(job_id, 
							DRMAA_JOBNAME_BUFFER-1, 
							jt, 
							error, 
							DRMAA_ERROR_STRING_BUFFER-1);
									
				if ( result != DRMAA_ERRNO_SUCCESS) 
				{
					f1 = fopen ("output.txt", "a");
					fprintf(f1,"drmaa_run_job() failed: %s\n", error);
					fclose(f1);
					return -1;    
				}
				
				f1 = fopen ("output.txt", "a");
				fprintf(f1,"Copy %d of Job %s successfully submitted\n",selfjob,job_id);
				fclose(f1);
				
				metajob[i][selfjob] = atoi(job_id);
				f1 = fopen ("output.txt", "a");
				fprintf(f1,"-> %d\n",metajob[i][selfjob]);
				fclose(f1);
				
				// Put on hold if not first level
	
				if (firstlevel == 0)
				{
					result = drmaa_control (job_id,
								DRMAA_CONTROL_HOLD,
								error, 
								DRMAA_ERROR_STRING_BUFFER-1);
				}
	
				if ( result != DRMAA_ERRNO_SUCCESS) 
				{
					f1 = fopen ("output.txt", "a");
					fprintf(f1,"Hold Job failed: %s\n", error);
					fclose(f1);
					return -1;    
				}
			}
		}
	}
	
	// If agglomeration, get the real jobs
       
	if (agglevel!=-1)
	{
		jobagg[0]=metajob[mtagg[0]][0];
		jobagg[1]=metajob[mtagg[1]][0];
		
		// aggnumjobs = jobagg[1] - jobagg[0] + 1;
		
		f1 = fopen ("output.txt", "a");
		fprintf (f1,"\nAgglomeration: Jobs %d to %d (%d jobs at first level)\n",jobagg[0],jobagg[1],aggnumjobs);
		fclose(f1);
	}
	//
	// Polling Time!
	//

	int status;
	int lastmt = 0;
	
	while(lastmt == 0)
	{
		
		// Get next job
		result = drmaa_wait(DRMAA_JOB_IDS_SESSION_ANY,
				job_id_out, 
				DRMAA_JOBNAME_BUFFER-1, 
				&stat, 
				DRMAA_TIMEOUT_WAIT_FOREVER, 
				&rusage, 
				error, 
				DRMAA_ERROR_STRING_BUFFER-1);
		// Which job?
		f1 = fopen ("output.txt", "a");
		fprintf(f1,"\n\n-> Job %s finished!",job_id_out);
		fclose(f1);
		
		// KNOW THE COORDS BUT NOT THE MT 
	
		// Which Metatask?
		for (i=0;i<mtnum;i++)
		{
			for (selfjob=0;selfjob<rellevel;selfjob++)
			{		
				// For each job
				
				//If not from critical level, just one copy
				
				if ((reltnum == 1) || ((mtcoord[i][0] == mtcoord[i][1]) || ((mtcoord[i][0] != mtcoord[i][1]) && (selfjob==0))))
				{
				
					if (metajob[i][selfjob] == atoi(job_id_out))
					{
						f1 = fopen ("output.txt", "a");
						fprintf(f1,"[Metatask %d ended]\n",i);
						fclose(f1);
						
						// Set metajob[i][selfjob] to -2 so it won't be asked to be killed
						metajob[i][selfjob] = -2;
												
						if (mtdone[i] != 1)
						{
							
							mtdone[i] = 1;
							mtdep[i][0] = -2;
							mtdep[i][1] = -2;
							
							// If last MT?
							if (i == (mtnum-1))
							{
								lastmt = 1; 
							}
						
							// Time to Launch dependencies and kill copies
							z = 0; // Used for level
							h = 0; // Used for that MT
							for (g=levels-1;g>0;g--)
							{
								for (j=z;j<levels-1;j++)
								{	
									if (((mtdone[mtdep[h][0]] == 1) && (mtdone[mtdep[h][1]] == 1)) && (mtreleased[h] == 0)) 
									{
										f1 = fopen ("output.txt", "a");
										fprintf (f1,"-> [R] MetaTask %d\n",h);
										fclose(f1);
										
										mtreleased[h] = 1;
										
										for (selfjob=0;selfjob<rellevel;selfjob++)
										{		
											// For each job
											
											//If not from critical level, just one copy
											
											if ((reltnum == 1) || ((mtcoord[h][0] == mtcoord[h][1]) || ((mtcoord[h][0] != mtcoord[h][1]) && (selfjob==0))))
											{
												// Release metatasks if possible
												sprintf(job_id,"%d",metajob[h][selfjob]);
												f1 = fopen ("output.txt", "a");
												fprintf(f1,"--> Will release %s (Metatask %d)\n", job_id,h);
												fclose(f1);
												
												// If agglomeration, check if will release agglomeration jobs
						
												if ((agglevel!=-1) && (h >= mtagg[0]))
												{
													aggnumjobs--;
													f1 = fopen ("output.txt", "a");
													fprintf(f1,"NO! Job is to be agglomerated (%d to go!)\n", aggnumjobs);
													fclose(f1);
													if (aggnumjobs == 0)
													{
														lastmt = 1;
														f1 = fopen ("output.txt", "a");
														fprintf(f1,"[I] Time for Agglomeration from Job %d to %d!\n", jobagg[0],jobagg[1]);
														fclose(f1);
														sprintf (aggcommand,"./agglomeration %d %d",jobagg[0],jobagg[1]);
														system(aggcommand);
													}
												}
												else							
												{
													// RELEASE!!!
													result = drmaa_control (job_id,
														DRMAA_CONTROL_RELEASE,
														error, 
														DRMAA_ERROR_STRING_BUFFER-1);
													
													if ( result != DRMAA_ERRNO_SUCCESS) 
													{
														f1 = fopen ("output.txt", "a");
														fprintf(f1,"Job Release failed: %s\n", error);
														fclose(f1);
														//return -1;    
													}
												}
											}
										}	
									}
	
		
									h++;
								}
// 								printf ("\n");
								z++;	
							}
							
						}

						// Kill shadows
						// metajob[i][selfjob]
						// se g as index

						
						for (g=0;g<rellevel;g++)
						{
							if (metajob[i][g] >= 0)
							{
								f1 = fopen ("output.txt", "a");
								fprintf(f1,"-> Will kill NOT Job %d\n", metajob[i][g]);
								fclose(f1);
								// KILL!!!
								sprintf(job_id,"%d",metajob[i][g]);
							}
						}
					}
				}
			}
		}

	}

    result = drmaa_exit (error, DRMAA_ERROR_STRING_BUFFER-1);

    if ( result != DRMAA_ERRNO_SUCCESS)
    {
      f1 = fopen ("output.txt", "a");	
      fprintf(f1,"drmaa_exit() failed: %s\n", error);
      fclose(f1);
      return -1;
    }

     
    return 0;
}

/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */

void setup_job_template( drmaa_job_template_t **jt, int coorda, int coordb, float c, int n)
{
	char        error[DRMAA_ERROR_STRING_BUFFER];
	int         rc;    
	char        cwd[DRMAA_ATTR_BUFFER];
	FILE *f1;
	
	f1 = fopen ("output.txt", "a");
	fprintf (f1,"-JT(%d,%d)-\n",coorda,coordb);
	fclose(f1);
	
	FILE *file;
	

	const char  *args[3] = {"-l", "-a", NULL};    
	
	rc = drmaa_allocate_job_template(jt, error, DRMAA_ERROR_STRING_BUFFER);
	
	if ( rc != DRMAA_ERRNO_SUCCESS) 
	{
		f1 = fopen ("output.txt", "a");
		fprintf(f1,"drmaa_allocate_job_template() failed: %s\n", error);
		fclose(f1);
		exit(-1);    
	}
		
	if ( getcwd(cwd, DRMAA_ATTR_BUFFER) == NULL )
	{
		perror("Error getting current working directory");
		exit(-1);    
	}
	
	rc = drmaa_set_attribute(*jt,
				DRMAA_WD,
				cwd,
				error,
				DRMAA_ERROR_STRING_BUFFER);
	
	if ( rc != DRMAA_ERRNO_SUCCESS )
	{
		f1 = fopen ("output.txt", "a");
		fprintf(f1,"Error setting job template attribute: %s\n",error);
		fclose(f1);
		exit(-1);    
	}    
	
	// Name of the Job Temaplate
	char	jtname[DRMAA_ATTR_BUFFER];
	

	sprintf (jtname,"jtcd%dvs%d",coorda,coordb);
	
	rc = drmaa_set_attribute(*jt,
				DRMAA_JOB_NAME,
				jtname,
				error,
				DRMAA_ERROR_STRING_BUFFER);
				
	// REQUIREMENTS
	rc = drmaa_set_attribute(*jt,
				DRMAA_GW_REQUIREMENTS,
				"(QUEUE_ACCESS = \"*biomed*\") | (HOSTNAME = \"*.ciemat.es\") | (HOSTNAME = \"*.ucm.es\")",
				error,
				DRMAA_ERROR_STRING_BUFFER);

	// Here Job Templates are different depending from where they are
	
	if (coorda == coordb + 1)
	{
		// Critical Path
		f1 = fopen ("output.txt", "a");
		fprintf(f1,"CP!\n");
		fclose(f1);
		
		// Create the BASH script
		char shname[DRMAA_ATTR_BUFFER];
		sprintf (shname,"jtcd.%d.vs.%d.sh",coorda,coordb);
		file = fopen (shname,"wt");
		fprintf (file,"#!/bin/bash\n\n");
		fprintf (file,"chmod +x cd-hit-2d\n");	
		fprintf (file,"chmod +x cd-hit\n");
		fprintf (file,"./cd-hit-2d -i protdb.div-%d-o -i2 protdb.div-%d -o protdb.div-%d.vs.%d -c %f -n %d\n",coordb,coorda,coorda,coordb,c,n);
		fprintf (file,"./cd-hit -i protdb.div-%d.vs.%d -o protdb.div-%d-o -c %f -n %d\n",coorda,coordb,coorda,c,n);	
		fclose (file);	
		// Now create the Job Template
		// Executable (the Script)
		const char* executable = shname;
		
		rc = drmaa_set_attribute(*jt, 
				DRMAA_REMOTE_COMMAND,
				executable, 
				error, 
				DRMAA_ERROR_STRING_BUFFER);
		
		// Input files		
		char file1[DRMAA_ATTR_BUFFER];
		sprintf (file1,"protdb.div-%d-o",coordb);
		char file2[DRMAA_ATTR_BUFFER];
		sprintf (file2,"protdb.div-%d",coorda);
		const char* inputfiles[5] = {"cd-hit-2d", "cd-hit", file1, file2, NULL};
		rc = drmaa_set_vector_attribute(*jt, 
				DRMAA_V_GW_INPUT_FILES, 
				inputfiles, 
				error,
				DRMAA_ERROR_STRING_BUFFER);
			if ( rc != DRMAA_ERRNO_SUCCESS )
			{
				f1 = fopen ("output.txt", "a");
				fprintf(f1,"Error: %s\n",error);
				fclose(f1);
				exit(-1);    
			}
				
		// Output files		
		sprintf (file1,"protdb.div-%d.vs.%d",coorda,coordb);
		
		sprintf (file2,"protdb.div-%d.vs.%d.clstr",coorda,coordb);
		char file3[DRMAA_ATTR_BUFFER];
		sprintf (file3,"protdb.div-%d-o",coorda);
		char file4[DRMAA_ATTR_BUFFER];
		sprintf (file4,"protdb.div-%d-o.clstr",coorda);
		
		const char* outputfiles[5] = {file1, file2, file3, file4, NULL};
		rc = drmaa_set_vector_attribute(*jt, 
				DRMAA_V_GW_OUTPUT_FILES, 
				outputfiles, 
				error,
				DRMAA_ERROR_STRING_BUFFER);		
				
	}
	else
	{
		// Not in Critical Path
		
		// Create the Job Template
		// Executable
		rc = drmaa_set_attribute(*jt, 
				DRMAA_REMOTE_COMMAND,
				"cd-hit-2d", 
				error, 
				DRMAA_ERROR_STRING_BUFFER);
		// Arguments
		char arg1[DRMAA_ATTR_BUFFER];
		sprintf (arg1,"-i protdb.div-%d-o",coordb);
		char arg2[DRMAA_ATTR_BUFFER];
		sprintf (arg2,"-i2 protdb.div-%d",coorda);
		char arg3[DRMAA_ATTR_BUFFER];
		sprintf (arg3,"-o protdb.div-%d.vs.%d",coorda,coordb);
		const char* args[4] = {arg1, arg2, arg3, NULL};
		rc = drmaa_set_vector_attribute(*jt,
                                  DRMAA_V_ARGV,
                                  args,
                                  error,
                                  DRMAA_ERROR_STRING_BUFFER-1);
		// Input files
		char file1[DRMAA_ATTR_BUFFER];
		sprintf (file1,"protdb.div-%d-o",coordb);
		char file2[DRMAA_ATTR_BUFFER];
		sprintf (file2,"protdb.div-%d",coorda);
		const char* inputfiles[3] = {file1, file2, NULL};
		rc = drmaa_set_vector_attribute(*jt, 
				DRMAA_V_GW_INPUT_FILES, 
				inputfiles, 
				error,
				DRMAA_ERROR_STRING_BUFFER);
		// Output files		
		sprintf (file1,"protdb.div-%d.vs.%d",coorda,coordb);
		sprintf (file2,"protdb.div-%d.vs.%d.clstr",coorda,coordb);
		const char* outputfiles[3] = {file1, file2, NULL};
		rc = drmaa_set_vector_attribute(*jt, 
				DRMAA_V_GW_OUTPUT_FILES, 
				outputfiles, 
				error,
				DRMAA_ERROR_STRING_BUFFER);
		
		
	}
	
	rc = drmaa_set_attribute(*jt, 
				DRMAA_OUTPUT_PATH, 
				":stdout."DRMAA_GW_JOB_ID, 
				error,
				DRMAA_ERROR_STRING_BUFFER);
	
	rc = drmaa_set_attribute(*jt, 
				DRMAA_ERROR_PATH, 
				":stderr."DRMAA_GW_JOB_ID, 
				error,
				DRMAA_ERROR_STRING_BUFFER);

	rc = drmaa_set_attribute(*jt,
                             DRMAA_GW_RESCHEDULE_ON_FAILURE,
                             "yes",
                             error,
                             DRMAA_ERROR_STRING_BUFFER);
}


